# -*- coding: utf-8 -*-
from xml.etree import cElementTree as eTree

import bustoken
import sysactions
from msgbus import TK_SYS_ACK, FM_PARAM
from mw_helper import show_message_dialog, show_numpad_dialog
from typing import Optional, Union

from .get_operator_level import get_user_level
from ..models import UserLevels

ERROR_MESSAGE = "$ERROR"
MAX_USER_ID_SIZE = 10


def get_authorization(
        pos_id,             # type: Union[int, str]
        min_level=None,     # type: Optional[int]
        model=None,         # type: Optional[eTree.Element]
        timeout=60000,      # type: Optional[int]
        display_title=None, # type: Optional[str]
        user_id=None,       # type: Optional[str]
):
    # type: (...) -> (bool, str)

    auth_type = sysactions.get_cfg(pos_id).key_value("authorizationType", "login").lower()

    ret_user_id = None
    success = False
    try:
        if auth_type == "fingerprint":
            success, ret_user_id = _login_by_fingerprint(
                pos_id=pos_id,
                min_level=min_level,
                user_id=user_id,
                timeout=timeout,
            )

            if success:
                return success, ret_user_id
            else:
                auth_type = "login"

        if auth_type == "login":
            success, ret_user_id = _login_by_numpad(
                pos_id=pos_id,
                min_level=min_level,
                model=model,
                display_title=display_title,
                user_id=user_id
            )
            if success:
                return success, ret_user_id

    except sysactions.AuthenticationFailed as ex:
        if str(ex) == "ACCESS_DENIED":
            show_message_dialog(pos_id, message="$ACCESS_DENIED")
        elif str(ex) == "ERROR_READING_FINGERPRINT":
            show_message_dialog(pos_id, message="ERROR_READING_FINGERPRINT")
        else:
            show_message_dialog(pos_id, message="$INVALID_USERNAME_OR_PASSWORD")

    finally:
        if success:
            sysactions.set_custom(pos_id, 'Last Manager ID', ret_user_id)
            sysactions.set_custom(pos_id, 'Authorization Level', get_user_level(ret_user_id))

    return False, None


def _login_by_fingerprint(
        pos_id,             # type: int
        min_level=None,     # type: Optional[int]
        user_id=None,       # type: Optional[str]
        timeout=60000,      # type: Optional[int]
):
    # type: (...) -> (bool, str)

    try:
        ret = sysactions.mbcontext.MB_EasySendMessage("FingerPrintReader{0}".format(pos_id), bustoken.TK_FPR_OK)
        if ret.token == TK_SYS_ACK:
            ret = sysactions.mbcontext.MB_EasySendMessage(
                "FingerPrintReader{0}".format(pos_id),
                bustoken.TK_FPR_AUTHORIZE_USER,
                format=FM_PARAM,
                data=pos_id,
                timeout=timeout,
            )

        if ret.token == bustoken.TK_FPR_TIMEOUT:
            return None, None
    except Exception as _:  # noqa
        return None, None

    if ret.token != TK_SYS_ACK:
        raise sysactions.AuthenticationFailed("ERROR_READING_FINGERPRINT")
    elif ret.data == "":
        raise sysactions.AuthenticationFailed("INVALID_USER_OR_PASSWORD")
    else:
        user_id = user_id or ret.data
        try:
            user_xml_str = sysactions.get_user_information(user_id)
        except Exception as _:  # noqa
            raise sysactions.AuthenticationFailed("ERROR_READING_FINGERPRINT")

        if user_xml_str is None:
            raise sysactions.AuthenticationFailed("ERROR_READING_FINGERPRINT")

    user_xml = eTree.XML(user_xml_str)
    user_element = user_xml.find("user")
    user_level = int(user_element.attrib["Level"])
    if user_level < min_level:
        raise sysactions.AuthenticationFailed("ACCESS_DENIED")

    return True, user_id


def _login_by_numpad(
        pos_id,             # type: int
        min_level=None,     # type: Optional[int]
        model=None,         # type: Optional[eTree.Element]
        display_title=None, # type: Optional[str]
        user_id=None,       # type: Optional[str]
):
    # type: (...) -> (bool, str)

    if display_title is None:
        display_title = _get_default_numpad_login_title(min_level, model, pos_id)

    if not user_id:
        user_id = show_numpad_dialog(pos_id, display_title, "$ENTER_USER_ID", mask='LOGIN')
        if user_id is None:
            return False, None
        if user_id == '':
            show_message_dialog(pos_id, message="$INVALID_USER")
            return False, None

    login_by_nfc = len(user_id) >= MAX_USER_ID_SIZE
    if login_by_nfc:
        user_info = sysactions.authenticate_user_nfc(nfc_uuid=user_id, min_level=min_level)
        user_id = user_info.get("UserId")
    else:
        password = show_numpad_dialog(pos_id, display_title, "$ENTER_PASSWORD|{}".format(user_id), "password")
        if password is None:
            return False, None

        sysactions.authenticate_user(user_id, password, min_level=min_level)

    sysactions.sys_log_info("Authentication succeeded for POS id: {}, user id: {}".format(pos_id, user_id))
    return True, user_id


def _get_default_numpad_login_title(
        min_level,      # type: int
        model,          # type: eTree.Element
        pos_id,         # type: int
):
    # type: (...) -> str

    model = model or sysactions.get_model(pos_id)
    user_level_name = UserLevels.get_name(min_level)
    if user_level_name is None:
        user_levels = UserLevels.list_values()
        for level in user_levels:
            if min_level <= level:
                user_level_name = UserLevels.get_name(level)
                break

    display_title = "$USER_AUTHENTICATION"
    if user_level_name is not None:
        level_name = sysactions.translate_message(model, user_level_name)
        display_title = "$USER_AUTHENTICATION_WITH_PARAMETER|{}".format(level_name)

    return display_title



# tin.py - functions for handling TINs
#
# Copyright (C) 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""TIN (U.S. Taxpayer Identification Number).

The Taxpayer Identification Number is used used for tax purposes in the
United States. A TIN may be:

* a Social Security Number (SSN)
* an Individual Taxpayer Identification Number (ITIN)
* an Employer Identification Number (EIN)
* a Preparer Tax Identification Number (PTIN)
* an Adoption Taxpayer Identification Number (ATIN)

>>> compact('123-45-6789')
'123456789'
>>> validate('123-45-6789')
'123456789'
>>> validate('07-3456789')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> guess_type('536-90-4399')
['ssn', 'atin']
>>> guess_type('04-2103594')
['ein']
>>> guess_type('042103594')
['ssn', 'ein', 'atin']
>>> format('042103594')
'042-10-3594'
>>> format('123-456')  # invalid numbers are not reformatted
'123-456'
"""

from stdnum.exceptions import *
from stdnum.us import atin, ein, itin, ptin, ssn
from stdnum.util import clean


_tin_modules = (ssn, itin, ein, ptin, atin)


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '-').strip()


def validate(number):
    """Check if the number is a valid TIN. This searches for the proper
    sub-type and validates using that."""
    for mod in _tin_modules:
        try:
            return mod.validate(number)
        except ValidationError:
            pass  # try next module
    # fallback
    raise InvalidFormat()


def is_valid(number):
    """Check if the number is a valid TIN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def guess_type(number):
    """Return a list of possible TIN types for which this number is
    valid.."""
    return [mod.__name__.rsplit('.', 1)[-1]
            for mod in _tin_modules
            if mod.is_valid(number)]


def format(number):
    """Reformat the number to the standard presentation format."""
    for mod in _tin_modules:
        if mod.is_valid(number) and hasattr(mod, 'format'):
            return mod.format(number)
    return number



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from collections.abc import Sequence
from typing import Any, Mapping, Optional, List, Callable, TypeVar, Dict

import backoff

from opentelemetry.sdk.util.instrumentation import InstrumentationScope
from opentelemetry.proto.common.v1.common_pb2 import (
    InstrumentationScope as PB2InstrumentationScope,
)
from opentelemetry.proto.resource.v1.resource_pb2 import (
    Resource as PB2Resource,
)
from opentelemetry.proto.common.v1.common_pb2 import AnyValue as PB2AnyValue
from opentelemetry.proto.common.v1.common_pb2 import KeyValue as PB2KeyValue
from opentelemetry.proto.common.v1.common_pb2 import (
    KeyValueList as PB2KeyValueList,
)
from opentelemetry.proto.common.v1.common_pb2 import (
    ArrayValue as PB2ArrayValue,
)
from opentelemetry.sdk.trace import Resource
from opentelemetry.util.types import Attributes


_logger = logging.getLogger(__name__)

_TypingResourceT = TypeVar("_TypingResourceT")
_ResourceDataT = TypeVar("_ResourceDataT")


def _encode_instrumentation_scope(
    instrumentation_scope: InstrumentationScope,
) -> PB2InstrumentationScope:
    if instrumentation_scope is None:
        return PB2InstrumentationScope()
    return PB2InstrumentationScope(
        name=instrumentation_scope.name,
        version=instrumentation_scope.version,
    )


def _encode_resource(resource: Resource) -> PB2Resource:
    return PB2Resource(attributes=_encode_attributes(resource.attributes))


def _encode_value(value: Any) -> PB2AnyValue:
    if isinstance(value, bool):
        return PB2AnyValue(bool_value=value)
    if isinstance(value, str):
        return PB2AnyValue(string_value=value)
    if isinstance(value, int):
        return PB2AnyValue(int_value=value)
    if isinstance(value, float):
        return PB2AnyValue(double_value=value)
    if isinstance(value, Sequence):
        return PB2AnyValue(
            array_value=PB2ArrayValue(values=[_encode_value(v) for v in value])
        )
    elif isinstance(value, Mapping):
        return PB2AnyValue(
            kvlist_value=PB2KeyValueList(
                values=[_encode_key_value(str(k), v) for k, v in value.items()]
            )
        )
    raise Exception(f"Invalid type {type(value)} of value {value}")


def _encode_key_value(key: str, value: Any) -> PB2KeyValue:
    return PB2KeyValue(key=key, value=_encode_value(value))


def _encode_span_id(span_id: int) -> bytes:
    return span_id.to_bytes(length=8, byteorder="big", signed=False)


def _encode_trace_id(trace_id: int) -> bytes:
    return trace_id.to_bytes(length=16, byteorder="big", signed=False)


def _encode_attributes(
    attributes: Attributes,
) -> Optional[List[PB2KeyValue]]:
    if attributes:
        pb2_attributes = []
        for key, value in attributes.items():
            try:
                pb2_attributes.append(_encode_key_value(key, value))
            except Exception as error:  # pylint: disable=broad-except
                _logger.exception(error)
    else:
        pb2_attributes = None
    return pb2_attributes


def _get_resource_data(
    sdk_resource_scope_data: Dict[Resource, _ResourceDataT],
    resource_class: Callable[..., _TypingResourceT],
    name: str,
) -> List[_TypingResourceT]:

    resource_data = []

    for (
        sdk_resource,
        scope_data,
    ) in sdk_resource_scope_data.items():
        collector_resource = PB2Resource(
            attributes=_encode_attributes(sdk_resource.attributes)
        )
        resource_data.append(
            resource_class(
                **{
                    "resource": collector_resource,
                    "scope_{}".format(name): scope_data.values(),
                }
            )
        )
    return resource_data


# Work around API change between backoff 1.x and 2.x. Since 2.0.0 the backoff
# wait generator API requires a first .send(None) before reading the backoff
# values from the generator.
_is_backoff_v2 = next(backoff.expo()) is None


def _create_exp_backoff_generator(*args, **kwargs):
    gen = backoff.expo(*args, **kwargs)
    if _is_backoff_v2:
        gen.send(None)
    return gen



# ASN.1 "useful" types
from pyasn1.type import char, tag

class ObjectDescriptor(char.GraphicString):
    tagSet = char.GraphicString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 7)
        )

class GeneralizedTime(char.VisibleString):
    tagSet = char.VisibleString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 24)
        )

class UTCTime(char.VisibleString):
    tagSet = char.VisibleString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 23)
        )



""" Find compiled module linking to Tcl / Tk libraries
"""
import sys

if sys.version_info.major > 2:
    from tkinter import _tkinter as tk
else:
    from Tkinter import tkinter as tk

if hasattr(sys, "pypy_find_executable"):
    # Tested with packages at https://bitbucket.org/pypy/pypy/downloads.
    # PyPies 1.6, 2.0 do not have tkinter built in.  PyPy3-2.3.1 gives an
    # OSError trying to import tkinter. Otherwise:
    try:  # PyPy 5.1, 4.0.0, 2.6.1, 2.6.0
        TKINTER_LIB = tk.tklib_cffi.__file__
    except AttributeError:
        # PyPy3 2.4, 2.1-beta1; PyPy 2.5.1, 2.5.0, 2.4.0, 2.3, 2.2, 2.1
        TKINTER_LIB = tk.tkffi.verifier.modulefilename
else:
    TKINTER_LIB = tk.__file__



# kennitala.py - functions for handling Icelandic identity codes
# coding: utf-8
#
# Copyright (C) 2015 Tuomas Toivonen
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Kennitala (Icelandic personal and organisation identity code).

Module for handling Icelandic personal and organisation identity codes
(kennitala).

>>> validate('450401-3150')  # organisation
'4504013150'
>>> validate('120174-3399')  # individual
'1201743399'
>>> validate('530575-0299')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('320174-3399')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> format('1201743399')
'120174-3399'
"""

import datetime
import re

from stdnum.exceptions import *
from stdnum.util import clean


# Icelandic personal and organisation identity codes are composed of
# date part, a dash, two random digits, a checksum, and a century
# indicator where '9' for 1900-1999 and '0' for 2000 and beyond. For
# organisations instead of birth date, the registration date is used,
# and number 4 is added to the first digit.
_kennitala_re = re.compile(
    r'^(?P<day>[01234567]\d)(?P<month>[01]\d)(?P<year>\d\d)'
    r'(?P<random>\d\d)(?P<control>\d)'
    r'(?P<century>[09])$')


def compact(number):
    """Convert the kennitala to the minimal representation. This
    strips surrounding whitespace and separation dash, and converts it
    to upper case."""
    return clean(number, '-').upper().strip()


def checksum(number):
    """Calculate the checksum."""
    weights = (3, 2, 7, 6, 5, 4, 3, 2, 1, 0)
    return sum(w * int(n) for w, n in zip(weights, number)) % 11


def validate(number):
    """Check if the number provided is a valid kennitala. It checks the
    format, whether a valid date is given and whether the check digit is
    correct."""
    number = compact(number)
    match = _kennitala_re.search(number)
    if not match:
        raise InvalidFormat()
    day = int(match.group('day'))
    month = int(match.group('month'))
    year = int(match.group('year'))
    if match.group('century') == '9':
        year += 1900
    else:
        year += 2000
    # check if birth date or registration data is valid
    try:
        if day >= 40:  # organisation
            datetime.date(year, month, day - 40)
        else:  # individual
            datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()
    # validate the checksum
    if checksum(number) != 0:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid HETU. It checks the format,
    whether a valid date is given and whether the check digit is correct."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return number[:6] + '-' + number[6:]



from interactor import BaseInteractor
from model.resync import ResyncRequest
from processor import BaseProcessor


class ResyncMessageProcessor(BaseProcessor):
    def __init__(self, interactor):
        # type: (BaseInteractor) -> None
        super(ResyncMessageProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data, event_type):
        params = data.split('\0')

        period = params[0] if params else None
        event_types = params[1] if len(params) > 1 else None
        last_row_id = params[2] if len(params) > 2 else None

        return ResyncRequest(period=period, event_types=event_types, last_row_id=last_row_id)

    def call_business(self, obj):
        self.interactor.execute(obj)



from abc import ABCMeta

from messagebus import AckMessage, NakMessage
from messageprocessorutil import UuidMessageProcessor
from production.command import ChangePathCommand
from production.manager import ProductionManager


class ChangePathProcessor(UuidMessageProcessor):
    __metaclass__ = ABCMeta

    def __init__(self, production_manager):
        # type: (ProductionManager) -> None
        super(ChangePathProcessor, self).__init__()
        self.production_manager = production_manager

    def call_business(self, request):
        change_path_command = ChangePathCommand(request.path, request.enabled)
        self.production_manager.handle_command(change_path_command)

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, AckMessage())

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, NakMessage())

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage())



import collections.abc
import functools


# from jaraco.functools 3.5
def pass_none(func):
    """
    Wrap func so it's not called if its first param is None

    >>> print_text = pass_none(print)
    >>> print_text('text')
    text
    >>> print_text(None)
    """

    @functools.wraps(func)
    def wrapper(param, *args, **kwargs):
        if param is not None:
            return func(param, *args, **kwargs)

    return wrapper


# from jaraco.functools 4.0
@functools.singledispatch
def _splat_inner(args, func):
    """Splat args to func."""
    return func(*args)


@_splat_inner.register
def _(args: collections.abc.Mapping, func):
    """Splat kargs to func as kwargs."""
    return func(**args)


def splat(func):
    """
    Wrap func to expect its parameters to be passed positionally in a tuple.

    Has a similar effect to that of ``itertools.starmap`` over
    simple ``map``.

    >>> import itertools, operator
    >>> pairs = [(-1, 1), (0, 2)]
    >>> _ = tuple(itertools.starmap(print, pairs))
    -1 1
    0 2
    >>> _ = tuple(map(splat(print), pairs))
    -1 1
    0 2

    The approach generalizes to other iterators that don't have a "star"
    equivalent, such as a "starfilter".

    >>> list(filter(splat(operator.add), pairs))
    [(0, 2)]

    Splat also accepts a mapping argument.

    >>> def is_nice(msg, code):
    ...     return "smile" in msg or code == 0
    >>> msgs = [
    ...     dict(msg='smile!', code=20),
    ...     dict(msg='error :(', code=1),
    ...     dict(msg='unknown', code=0),
    ... ]
    >>> for msg in filter(splat(is_nice), msgs):
    ...     print(msg)
    {'msg': 'smile!', 'code': 20}
    {'msg': 'unknown', 'code': 0}
    """
    return functools.wraps(func)(functools.partial(_splat_inner, func=func))



import json

from typing import (
    Any,
    List,
    Dict,
)


def set_table_loyalty_to_order(
    table,          # type: Dict[str, Any]
    properties,     # type: Dict[str, Any]
):
    # type: (...) -> None

    loyalty_id = (table.get("properties", {}) or {}).get("LOYALTY_TABLE_ID", "{}")
    customers_info = (table.get("properties", {}) or {}).get("LOYALTY_TABLE_CUSTOMER_INFO", "{}")
    all_loyalty_benefits = (table.get("properties", {}) or {}).get("LOYALTY_TABLE_CUSTOMER_BENEFITS", "{}")
    loyalty_benefits_to_apply = (table.get("properties", {}) or {}).get("LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL", "{}")
    loyalty_id = json.loads(s=loyalty_id)
    customers_info = json.loads(s=customers_info)
    all_loyalty_benefits = json.loads(s=all_loyalty_benefits)
    loyalty_benefits_to_apply = json.loads(s=loyalty_benefits_to_apply)

    joined_all_loyalty_benefits = []
    joined_loyalty_benefits_to_apply = []

    for key, benefits_ids in loyalty_benefits_to_apply.items():
        if key == "from_seat":
            continue
        current_benefits_ids = benefits_ids[:]
        if key != "table":
            current_benefits_ids = [
                "{}_{}".format(
                    c,
                    key,
                ) for c in current_benefits_ids
            ]
        joined_loyalty_benefits_to_apply.extend(current_benefits_ids)

    for key, benefits in all_loyalty_benefits.items():
        current_benefits = benefits[:]
        if key != "table":
            current_benefits = _set_seat_in_benefits(
                benefits=current_benefits,
                seat=key,
            )
        joined_all_loyalty_benefits.extend(current_benefits)

    if joined_all_loyalty_benefits:
        properties["LOYALTY_CUSTOMER_BENEFITS"] = json.dumps(obj=joined_all_loyalty_benefits)

    if joined_loyalty_benefits_to_apply:
        properties["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"] = json.dumps(obj=joined_loyalty_benefits_to_apply)

    loyalty_table = loyalty_id.get("table", "")
    if loyalty_table:
        properties["LOYALTY_ID"] = str(loyalty_table)

    customer_info_table = customers_info.get("table", {})
    if customer_info_table:
        properties["LOYALTY_CUSTOMER_INFO"] = json.dumps(obj=customer_info_table)


def _set_seat_in_benefits(
    benefits,   # type: List[Dict]
    seat,       # type: str
):
    # type: (...) -> List[Dict]

    seat = int(seat) if seat.isdigit() else None

    if not seat:
        return []
    seat_dict = dict(seat=seat)

    for benefit in benefits:
        benefit.update(dict(id="{}_{}".format(
            benefit.get("id", ""),
            seat,
        )))

        benefit.update(dict(userDescription="$LOYALTY_USER_DESCRIPTION_COUPON_WITH_SEAT|{}|{}".format(
            benefit.get("userDescription", "").encode("utf-8"),
            seat,
        )))

        action_keys = benefit.get("actions", {}).keys()
        for key in action_keys:
            action = benefit.get("actions", {}).get(key, "")
            if "discountPercentageOnTotal" in action or "discountAmountOnTotal" in action:
                action.update(seat_dict)
                continue
            action_type = action.keys()
            action_type = action_type[0] if len(action_type) == 1 else None
            if not action_type:
                continue
            for item in action.get(action_type, []):
                item.update(seat_dict)

    return benefits




import logging
from logging import Handler, LogRecord

from systools import sys_log_error, sys_log_debug, sys_log_info, sys_log_warning


class HvLogFormatter(logging.Formatter):

    def __init__(self, fmt=None, datefmt=None, style='%'):
        if fmt is None:
            fmt = "%(filename)s:%(lineno)s:%(funcName)s - %(message)s"
        super(HvLogFormatter, self).__init__(fmt, datefmt)


class HvLogHandler(Handler):

    def __init__(
        self,                     # type: HvLogHandler
        level_no=logging.ERROR,   # type: int
    ):
        super(HvLogHandler, self).__init__(level_no)
        self.log_func = {
            logging.DEBUG: sys_log_debug,
            logging.INFO: sys_log_info,
            logging.WARNING: sys_log_warning,
            logging.ERROR: sys_log_error,
        }
        self.setFormatter(HvLogFormatter())

    def emit(
        self,    # type: HvLogHandler
        record,  # type: LogRecord
    ):
        # type: (...) -> None

        log_func = sys_log_debug
        if record.levelno in self.log_func:
            log_func = self.log_func[record.levelno]
        log_func(self.format(record))



import os
from dataclasses import dataclass
from typing import Sequence


@dataclass(frozen=True)
class AlembicDatabaseInfo(object):
    name: str
    conn_strings: Sequence[str]
    alembic_path: str
    versions_dir: str

    def __init__(self, name: str, conn_strings: Sequence[str]):
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "conn_strings", conn_strings)
        object.__setattr__(self, "alembic_path", self._get_alembic_path())
        object.__setattr__(self, "versions_dir", self._get_version_dir())

    def _get_alembic_path(self) -> str:
        return self._get_parent_alembic_path(os.path.abspath(__file__))

    def _get_version_dir(self):
        name_dir = os.path.join(self.alembic_path, self.name)
        if os.path.isdir(name_dir):
            return name_dir
        return os.path.join(self.alembic_path, "versions")

    def _get_parent_alembic_path(self, current_dir) -> str:
        alembic_path = os.path.join(current_dir, "alembic")
        if self._is_valid_alembic_dir(alembic_path=alembic_path):
            return alembic_path
        parent_dir = os.path.dirname(current_dir)
        if parent_dir == current_dir:
            raise RuntimeError("Could not find alembic path")
        return self._get_parent_alembic_path(parent_dir)

    @staticmethod
    def _is_valid_alembic_dir(alembic_path):
        if not os.path.exists(alembic_path) or not os.path.isdir(alembic_path):
            return False

        alembic_ini = os.path.join(alembic_path, "alembic.ini")
        return os.path.exists(alembic_ini) and os.path.isfile(alembic_ini)



from production.box import SorterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class SorterBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(SorterBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository
        self.loggers = loggers

    def build(self, config):
        # type: (BoxConfiguration) -> SorterBox
        if u"SortOrder" not in config.extra_config:
            raise Exception("Sequencer without Sorters: {}".format(config.name))

        return SorterBox(
            config.name,
            config.sons,
            self.product_repository,
            self.get_extra(config, "SortOrder", None),
            self.get_logger(config)
        )



# siret.py - functions for handling French SIRET numbers
# coding: utf-8
#
# Copyright (C) 2016 Yoann Aubineau
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""SIRET (a French company establishment identification number).

The SIRET (Système d'Identification du Répertoire des ETablissements)
is a 14 digit number used to identify French companies' establishments
and facilities. The Luhn checksum is used to validate the numbers (except
for La Poste).

More information:

* https://fr.wikipedia.org/wiki/Système_d'identification_du_répertoire_des_établissements

>>> validate('73282932000074')
'73282932000074'
>>> validate('73282932000079')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> to_siren('732 829 320 00074')
'732 829 320'
>>> to_siren('73282932000074')
'732829320'
>>> to_tva('732 829 320 00074')
'44 732 829 320'
>>> to_tva('73282932000074')
'44732829320'
>>> format('73282932000074')
'732 829 320 00074'
"""

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.fr import siren
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' .').strip()


def validate(number):
    """Check if the number is a valid SIRET. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 14:
        raise InvalidLength()
    # La Poste SIRET (except the head office) do not use the Luhn checksum
    # but the sum of digits must be a multiple of 5
    if number.startswith('356000000') and number != '35600000000048':
        if sum(map(int, number)) % 5 != 0:
            raise InvalidChecksum()
    else:
        luhn.validate(number)
    siren.validate(number[:9])
    return number


def is_valid(number):
    """Check if the number is a valid SIRET."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def to_siren(number):
    """Convert the SIRET number to a SIREN number.

    The SIREN number is the 9 first digits of the SIRET number.
    """
    _siren = []
    digit_count = 0
    for char in number:
        if digit_count < 9:
            _siren.append(char)
            if isdigits(char):
                digit_count += 1
    return ''.join(_siren)


def to_tva(number):
    """Convert the SIRET number to a TVA number.

    The TVA number is built from the SIREN number, prepended by two extra
    error checking digits.
    """
    return siren.to_tva(to_siren(number))


def format(number, separator=' '):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return separator.join((number[0:3], number[3:6], number[6:9], number[9:]))



from typing import Callable, Any
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta
from threading import Lock, Thread


class CacheSource(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_new_object(self):
        raise NotImplementedError()


class CacheExpiration(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def is_expired(self):
        raise NotImplementedError()

    @abstractmethod
    def renew(self):
        raise NotImplementedError()


class CacheManager:
    def __init__(self, cache_expiration, cache_source=None, new_object_func=None, threaded=False):
        # type: (CacheExpiration, CacheSource, Callable[[], Any], bool) -> None
        self.cache_expiration = cache_expiration
        self.cache_source = cache_source
        self.new_object_func = new_object_func
        self.threaded = threaded

        self.retrieving_object = False
        self.cached_object = None
        self.renew_lock = Lock()

    def get_cached_object(self, new_object_func=None):
        if (not self.cache_expiration.is_expired() or self.retrieving_object) and self.cached_object is not None:
            return self.cached_object

        with self.renew_lock:
            if (not self.cache_expiration.is_expired() or self.retrieving_object) and self.cached_object is not None:
                return self.cached_object

            self.retrieving_object = True

            if self.threaded:
                renew_thread = Thread(target=self._renew_cache, args=new_object_func)
                renew_thread.daemon = True
                renew_thread.start()
            else:
                self._renew_cache(new_object_func)

            return self.cached_object

    def _renew_cache(self, new_object_func):
        try:
            if new_object_func is not None:
                self.cached_object = new_object_func()
            elif self.new_object_func is not None:
                self.cached_object = self.new_object_func()
            else:
                self.cached_object = self.cache_source.get_new_object()

            self.cache_expiration.renew()
        finally:
            self.retrieving_object = False


class PeriodBasedExpiration(CacheExpiration):
    def __init__(self, period_in_minutes):
        self.periodInMinutes = period_in_minutes
        self.expiration_date = datetime.now() - timedelta(days=1)

    def is_expired(self):
        return datetime.now() > self.expiration_date

    def renew(self):
        self.expiration_date = datetime.now() + timedelta(minutes=self.periodInMinutes)


class InfiniteExpiration(CacheExpiration):
    def __init__(self):
        super(CacheExpiration, self).__init__()

    def is_expired(self):
        return False

    def renew(self):
        pass



# nipt.py - functions for handling Albanian tax numbers
# coding: utf-8
#
# Copyright (C) 2008-2011 Cédric Krier
# Copyright (C) 2008-2011 B2CK
# Copyright (C) 2015-2023 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIPT, NUIS (Numri i Identifikimit për Personin e Tatueshëm, Albanian tax number).

The Albanian NIPT is a 10-digit number with the first and last character
being letters. The number is assigned to individuals and organisations for
tax purposes.

The first letter indicates the decade the number was assigned or date birth
date for individuals, followed by a digit for the year. The next two digits
contain the month (and gender for individuals and region for organisations)
followed by two digits for the day of the month. The remainder is a serial
followed by a check letter (check digit algorithm unknown).

More information:

* https://www.tatime.gov.al/eng/c/4/103/business-lifecycle

>>> validate('AL J 91402501 L')
'J91402501L'
>>> validate('K22218003V')
'K22218003V'
>>> validate('(AL) J 91402501')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('Z 22218003 V')
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

import re

from stdnum.exceptions import *
from stdnum.util import clean


# regular expression for matching number
_nipt_re = re.compile(r'^[A-M][0-9]{8}[A-Z]$')


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' ').upper().strip()
    if number.startswith('AL'):
        number = number[2:]
    if number.startswith('(AL)'):
        number = number[4:]
    return number


def validate(number):
    """Check if the number is a valid VAT number. This checks the length and
    formatting."""
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength()
    if not _nipt_re.match(number):
        raise InvalidFormat()
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from common import InfAdicBuilderCommon
from ibpttaxprocessor import IbptTaxProcessor
from pos_model import Order, XmlTenderTypeFromPOSMapper

from _SatXmlPartBuilder import SatXmlPartBuilder
from satbuilder import ContextKeys


class InfAdicBuilder(SatXmlPartBuilder):
    def __init__(self, customer_text_field, tax_builder):
        # type: (str, IbptTaxProcessor) -> None

        self.customer_text_field = customer_text_field
        self.tax_builder = tax_builder

    def build_xml(self, order, context):
        # type: (Order, ContextKeys) -> unicode

        return InfAdicBuilderCommon(order, self.customer_text_field, self.tax_builder).build_common()



from _TranspBuilder import TranspBuilder
from _ContextKeys import ContextKeys
from _DestinatarioBuilder import DestinatarioBuilder
from _DetBuilder import DetBuilder
from _DetTipBuilder import DetTipBuilder
from _DetFeeBuilder import DetFeeBuilder
from _DetListBuilder import DetListBuilder
from _EmitBuilder import EmitBuilder
from _IdeBuilder import IdeBuilder
from _InfNfeBuilder import InfNfeBuilder
from _InfNFeSuplBuilder import InfNFeSuplBuilder
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _NfeBuilder import NfeBuilder
from _PagBuilder import PagBuilder
from _TotalBuilder import TotalBuilder
from _TranspBuilder import TranspBuilder
from _NfceNamespaces import NfceNamespaces
from _TaxBuilder import TaxBuilder, TaxBuilderLucroReal, TaxBuilderSimplesNacional
from _InfRespTecBuilder import InfRespTecBuilder
from ._InfIntermedBuilder import InfIntermedBuilder
from _InfAdicBuilder import InfAdicBuilder



# coding: utf-8

"""
ASN.1 type classes for certificate signing requests (CSR). Exports the
following items:

 - CertificatationRequest()

Other type classes are defined that help compose the types listed above.
"""

from __future__ import unicode_literals, division, absolute_import, print_function

from .algos import SignedDigestAlgorithm
from .core import (
    Any,
    Integer,
    ObjectIdentifier,
    OctetBitString,
    Sequence,
    SetOf,
)
from .keys import PublicKeyInfo
from .x509 import DirectoryString, Extensions, Name


# The structures in this file are taken from https://tools.ietf.org/html/rfc2986
# and https://tools.ietf.org/html/rfc2985


class Version(Integer):
    _map = {
        0: 'v1',
    }


class CSRAttributeType(ObjectIdentifier):
    _map = {
        '1.2.840.113549.1.9.7': 'challenge_password',
        '1.2.840.113549.1.9.9': 'extended_certificate_attributes',
        '1.2.840.113549.1.9.14': 'extension_request',
    }


class SetOfDirectoryString(SetOf):
    _child_spec = DirectoryString


class Attribute(Sequence):
    _fields = [
        ('type', ObjectIdentifier),
        ('values', SetOf, {'spec': Any}),
    ]


class SetOfAttributes(SetOf):
    _child_spec = Attribute


class SetOfExtensions(SetOf):
    _child_spec = Extensions


class CRIAttribute(Sequence):
    _fields = [
        ('type', CSRAttributeType),
        ('values', Any),
    ]

    _oid_pair = ('type', 'values')
    _oid_specs = {
        'challenge_password': SetOfDirectoryString,
        'extended_certificate_attributes': SetOfAttributes,
        'extension_request': SetOfExtensions,
    }


class CRIAttributes(SetOf):
    _child_spec = CRIAttribute


class CertificationRequestInfo(Sequence):
    _fields = [
        ('version', Version),
        ('subject', Name),
        ('subject_pk_info', PublicKeyInfo),
        ('attributes', CRIAttributes, {'implicit': 0, 'optional': True}),
    ]


class CertificationRequest(Sequence):
    _fields = [
        ('certification_request_info', CertificationRequestInfo),
        ('signature_algorithm', SignedDigestAlgorithm),
        ('signature', OctetBitString),
    ]



class CloseOperator(object):
    def __init__(self, pos_id, values, envelope, session, payments, manager_id, period):
        self.pos_id = pos_id
        self.values = values
        self.envelope = envelope
        self.session = session
        self.payments = payments
        self.manager_id = manager_id
        self.period = period



# -*- coding: utf-8 -*-
import warnings
from datetime import tzinfo

from . import _compat
from ._exceptions import (
    AmbiguousTimeError,
    NonExistentTimeError,
    PytzUsageWarning,
    UnknownTimeZoneError,
    get_exception,
)

IS_DST_SENTINEL = object()
KEY_SENTINEL = object()


def timezone(key, _cache={}):
    """Builds an IANA database time zone shim.

    This is the equivalent of ``pytz.timezone``.

    :param key:
        A valid key from the IANA time zone database.

    :raises UnknownTimeZoneError:
        If an unknown value is passed, this will raise an exception that can be
        caught by :exc:`pytz_deprecation_shim.UnknownTimeZoneError` or
        ``pytz.UnknownTimeZoneError``. Like
        :exc:`zoneinfo.ZoneInfoNotFoundError`, both of those are subclasses of
        :exc:`KeyError`.
    """
    instance = _cache.get(key, None)
    if instance is None:
        if len(key) == 3 and key.lower() == "utc":
            instance = _cache.setdefault(key, UTC)
        else:
            try:
                zone = _compat.get_timezone(key)
            except KeyError:
                raise get_exception(UnknownTimeZoneError, key)
            instance = _cache.setdefault(key, wrap_zone(zone, key=key))

    return instance


def fixed_offset_timezone(offset, _cache={}):
    """Builds a fixed offset time zone shim.

    This is the equivalent of ``pytz.FixedOffset``. An alias is available as
    ``pytz_deprecation_shim.FixedOffset`` as well.

    :param offset:
        A fixed offset from UTC, in minutes. This must be in the range ``-1439
        <= offset <= 1439``.

    :raises ValueError:
        For offsets whose absolute value is greater than or equal to 24 hours.

    :return:
        A shim time zone.
    """
    if not (-1440 < offset < 1440):
        raise ValueError("absolute offset is too large", offset)

    instance = _cache.get(offset, None)
    if instance is None:
        if offset == 0:
            instance = _cache.setdefault(offset, UTC)
        else:
            zone = _compat.get_fixed_offset_zone(offset)
            instance = _cache.setdefault(offset, wrap_zone(zone, key=None))

    return instance


def build_tzinfo(zone, fp):
    """Builds a shim object from a TZif file.

    This is a shim for ``pytz.build_tzinfo``. Given a value to use as the zone
    IANA key and a file-like object containing a valid TZif file (i.e.
    conforming to :rfc:`8536`), this builds a time zone object and wraps it in
    a shim class.

    The argument names are chosen to match those in ``pytz.build_tzinfo``.

    :param zone:
        A string to be used as the time zone object's IANA key.

    :param fp:
        A readable file-like object emitting bytes, pointing to a valid TZif
        file.

    :return:
        A shim time zone.
    """
    zone_file = _compat.get_timezone_file(fp)

    return wrap_zone(zone_file, key=zone)


def wrap_zone(tz, key=KEY_SENTINEL, _cache={}):
    """Wrap an existing time zone object in a shim class.

    This is likely to be useful if you would like to work internally with
    non-``pytz`` zones, but you expose an interface to callers relying on
    ``pytz``'s interface. It may also be useful for passing non-``pytz`` zones
    to libraries expecting to use ``pytz``'s interface.

    :param tz:
        A :pep:`495`-compatible time zone, such as those provided by
        :mod:`dateutil.tz` or :mod:`zoneinfo`.

    :param key:
        The value for the IANA time zone key. This is optional for ``zoneinfo``
        zones, but required for ``dateutil.tz`` zones.

    :return:
        A shim time zone.
    """
    if key is KEY_SENTINEL:
        key = getattr(tz, "key", KEY_SENTINEL)

    if key is KEY_SENTINEL:
        raise TypeError(
            "The `key` argument is required when wrapping zones that do not "
            + "have a `key` attribute."
        )

    instance = _cache.get((id(tz), key), None)
    if instance is None:
        instance = _cache.setdefault((id(tz), key), _PytzShimTimezone(tz, key))

    return instance


class _PytzShimTimezone(tzinfo):
    # Add instance variables for _zone and _key because this will make error
    # reporting with partially-initialized _BasePytzShimTimezone objects
    # work better.
    _zone = None
    _key = None

    def __init__(self, zone, key):
        self._key = key
        self._zone = zone

    def utcoffset(self, dt):
        return self._zone.utcoffset(dt)

    def dst(self, dt):
        return self._zone.dst(dt)

    def tzname(self, dt):
        return self._zone.tzname(dt)

    def fromutc(self, dt):
        # The default fromutc implementation only works if tzinfo is "self"
        dt_base = dt.replace(tzinfo=self._zone)
        dt_out = self._zone.fromutc(dt_base)

        return dt_out.replace(tzinfo=self)

    def __str__(self):
        if self._key is not None:
            return str(self._key)
        else:
            return repr(self)

    def __repr__(self):
        return "%s(%s, %s)" % (
            self.__class__.__name__,
            repr(self._zone),
            repr(self._key),
        )

    def unwrap_shim(self):
        """Returns the underlying class that the shim is a wrapper for.

        This is a shim-specific method equivalent to
        :func:`pytz_deprecation_shim.helpers.upgrade_tzinfo`. It is provided as
        a method to allow end-users to upgrade shim timezones without requiring
        an explicit dependency on ``pytz_deprecation_shim``, e.g.:

        .. code-block:: python

            if getattr(tz, "unwrap_shim", None) is None:
                tz = tz.unwrap_shim()
        """
        return self._zone

    @property
    def zone(self):
        warnings.warn(
            "The zone attribute is specific to pytz's interface; "
            + "please migrate to a new time zone provider. "
            + "For more details on how to do so, see %s"
            % PYTZ_MIGRATION_GUIDE_URL,
            PytzUsageWarning,
            stacklevel=2,
        )

        return self._key

    def localize(self, dt, is_dst=IS_DST_SENTINEL):
        warnings.warn(
            "The localize method is no longer necessary, as this "
            + "time zone supports the fold attribute (PEP 495). "
            + "For more details on migrating to a PEP 495-compliant "
            + "implementation, see %s" % PYTZ_MIGRATION_GUIDE_URL,
            PytzUsageWarning,
            stacklevel=2,
        )

        if dt.tzinfo is not None:
            raise ValueError("Not naive datetime (tzinfo is already set)")

        dt_out = dt.replace(tzinfo=self)

        if is_dst is IS_DST_SENTINEL:
            return dt_out

        dt_ambiguous = _compat.is_ambiguous(dt_out)
        dt_imaginary = (
            _compat.is_imaginary(dt_out) if not dt_ambiguous else False
        )

        if is_dst is None:
            if dt_imaginary:
                raise get_exception(
                    NonExistentTimeError, dt.replace(tzinfo=None)
                )

            if dt_ambiguous:
                raise get_exception(AmbiguousTimeError, dt.replace(tzinfo=None))

        elif dt_ambiguous or dt_imaginary:
            # Start by normalizing the folds; dt_out may have fold=0 or fold=1,
            # but we need to know the DST offset on both sides anyway, so we
            # will get one datetime representing each side of the fold, then
            # decide which one we're going to return.
            if _compat.get_fold(dt_out):
                dt_enfolded = dt_out
                dt_out = _compat.enfold(dt_out, fold=0)
            else:
                dt_enfolded = _compat.enfold(dt_out, fold=1)

            # Now we want to decide whether the fold=0 or fold=1 represents
            # what pytz would return for `is_dst=True`
            enfolded_dst = bool(dt_enfolded.dst())
            if bool(dt_out.dst()) == enfolded_dst:
                # If this is not a transition between standard time and
                # daylight saving time, pytz will consider the larger offset
                # the DST offset.
                enfolded_dst = dt_enfolded.utcoffset() > dt_out.utcoffset()

            # The default we've established is that dt_out is fold=0; swap it
            # for the fold=1 datetime if is_dst == True and the enfolded side
            # is DST or if is_dst == False and the enfolded side is *not* DST.
            if is_dst == enfolded_dst:
                dt_out = dt_enfolded

        return dt_out

    def normalize(self, dt):
        warnings.warn(
            "The normalize method is no longer necessary, as this "
            + "time zone supports the fold attribute (PEP 495). "
            + "For more details on migrating to a PEP 495-compliant "
            + "implementation, see %s" % PYTZ_MIGRATION_GUIDE_URL,
            PytzUsageWarning,
            stacklevel=2,
        )

        if dt.tzinfo is None:
            raise ValueError("Naive time - no tzinfo set")

        if dt.tzinfo is self:
            return dt

        return dt.astimezone(self)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo=None):
        return self

    def __reduce__(self):
        return wrap_zone, (self._zone, self._key)


UTC = wrap_zone(_compat.UTC, "UTC")
PYTZ_MIGRATION_GUIDE_URL = (
    "https://pytz-deprecation-shim.readthedocs.io/en/latest/migration.html"
)



import logging
import threading
from logging import Logger
from typing import Optional, List, Type, Dict

from messageprocessor import MessageProcessorCallback

default_logger = logging.getLogger("MessageProcessor.LoggingProcessorCallback")


class LoggingProcessorCallback(MessageProcessorCallback):
    def __init__(self, logger, ignored_exceptions=None, log_level_exceptions=None):
        # type: (Logger, Optional[List[Type[Exception]]], Optional[Dict[Type[BaseException], int]]) -> None
        self.logger = logger if logger is not None else default_logger
        self.ignored_exceptions = {}
        if ignored_exceptions is not None:
            for ex in ignored_exceptions:
                self.ignored_exceptions[ex] = ex
        self.local_storage = threading.local()
        self.log_level_exceptions = log_level_exceptions or {}

    def process_begun(self, event_name, request_id, data):
        # type: (str, str, bytes) -> None
        self.local_storage.original_thread_name = threading.current_thread().name
        self._set_thread_name(event_name, request_id)
        self.logger.debug(b"Start handling. Data: {}" + data)

    def input_parsed(self, event_name, request_id, input_obj):
        self._set_thread_name(event_name, request_id)
        self.logger.debug(u"Input correctly parsed")

    def business_called(self, event_name, request_id, input_obj, output_obj):
        return

    def process_finished(self, event_name, request_id, input_obj, output_obj):
        self.logger.info(u"Successfully processed")
        self._restore_thread_name()

    def parse_exception(self, event_name, request_id, data, exception):
        self._log_exception(exception, u"Parse exception handling")
        self._restore_thread_name()

    def unhandled_exception(self, event_name, request_id, input_obj, exception):
        self._log_exception(exception, u"Unhandled exception handling")
        self._restore_thread_name()

    @staticmethod
    def _set_thread_name(event_name, request_id):
        threading.current_thread().name = "MessageProcessorThread${}${}".format(event_name, request_id)

    def _restore_thread_name(self):
        if hasattr(self.local_storage, "original_thread_name"):
            threading.current_thread().name = self.local_storage.original_thread_name

    def _log_exception(self, exception, message):
        type_exception = type(exception)

        if type_exception in self.ignored_exceptions:
            return

        level = self.log_level_exceptions.get(type_exception, logging.ERROR)
        self.logger.log(level, message, exc_info=True)
